-------------------------------------------------------------------

             Solving the Snake Cube Puzzle in Haskell

                          MARK P. JONES
                  Department of Computer Science
          Portland State University, Portland, Oregon, USA
                         mpj@cs.pdx.edu

-------------------------------------------------------------------
This file contains Haskell source code for the program described in
the above paper.  Please refer to the paper, available online from
http://web.cecs.pdx.edu/~mpj/snakecube/ for further explanation of
the code that is presented here.
-------------------------------------------------------------------

> module Snake where

-------------------------------------------------------------------
Constructing the Snake Cube:

The structure of the standard snake cube puzzle is described by the
following list of integers:

> snake :: [Int]
> snake  = [ 3, 2, 2, 3, 2, 3, 2, 2, 3, 3, 2, 2, 2, 3, 3, 3, 3 ]

There are sanity checks that we can run on the puzzles we define.
Using only the list of integers above, we can confirm that there
are 27 small cubes:

> sanityCheck :: Bool
> sanityCheck  = sum snake - (length snake - 1) == 27

We can also define a sanity check for the more general class of
puzzles that will be defined below.

> sanity      :: Puzzle -> Int
> sanity p     = sum (sections p) - (length (sections p) - 1)

-------------------------------------------------------------------
Moving in to Three Dimensions:

> type Position  = (Int, Int, Int)
> type Direction = (Int, Int, Int)

> inCube            :: Int -> Position -> Bool
> inCube n (x, y, z) = valid x && valid y && valid z
>                      where valid i = 1<=i && i<=n

-------------------------------------------------------------------
Describing Solutions:

> type Solution = [Section]
> type Section  = [Position]

-------------------------------------------------------------------
Building Sections:

> section                  :: Position -> Direction -> Int -> Section
> section start (u,v,w) len = reverse (tail (take len pieces))
>  where pieces = iterate (\ (x, y, z) -> (x+u, y+v, z+w)) start

-------------------------------------------------------------------
Changing Directions:

> newDirs    :: Direction -> [Direction]
> newDirs dir = [ rotate (sign dir) | rotate <- [left,right],
>                                     sign   <- [id, inv] ]
>  where left, right, inv :: Direction -> Direction
>        left  (x, y, z)   = ( y,  z,  x)
>        right (x, y, z)   = ( z,  x,  y)
>        inv   (x, y, z)   = (-x, -y, -z)

-------------------------------------------------------------------
Describing Complete Puzzles:

> data Puzzle = Puzzle { sections :: [Int],
>                        valid    :: Position -> Bool,
>                        initSoln :: Solution,
>                        initDir  :: Direction }

The standard snake cube puzzle:

> standard  :: Puzzle
> standard   = Puzzle { valid    = inCube 3,
>                       initSoln = [[(1,1,1)]],
>                       initDir  = (0,0,1),
>                       sections = snake }

The "mean green" and "ornery orange" from Creative Crafthouse:

> meanGreen :: Puzzle
> meanGreen  = standard { sections = [ 3, 3, 2, 3, 2, 3, 2, 2,
>                                      2, 3, 3, 3, 2, 3, 3, 3 ] }

> ornery    :: Puzzle
> ornery     = standard { sections = [ 3, 2, 2, 3, 2, 2, 3, 2, 2, 2,
>                                      2, 2, 2, 2, 3, 2, 2, 2, 3, 3] }

The two snake configurations illustrated in Allen F. Dreyer's US
Patent 3,222,072, awarded December 7, 1965:

> dreyerFig2 :: Puzzle
> dreyerFig2  = standard { sections = [3,3,3,2,3,2,3,2,2,2,3,2,3,2,2,2,2,2,2] }

> dreyerFig3 :: Puzzle
> dreyerFig3  = standard { sections = [2,2,2,2,2,2,3,3,2,2,2,2,2,2,2,2,2,2,3,2,2,2,2] }

The 4x4x4 king snake, including some variations:

> king  :: Puzzle
> king   = standard {
>            valid    = inCube 4,
>            sections = [ 3, 2, 3, 2, 2, 4, 2, 3, 2, 3, 2, 3, 2, 2, 2,
>                         2, 2, 2, 2, 2, 3, 3, 2, 2, 2, 2, 2, 3, 4, 2,
>                         2, 2, 4, 2, 3, 2, 2, 2, 2, 2, 2, 2, 2, 2, 4, 2 ] }

> king1, king2 :: Puzzle
> king1         = king { initSoln = [[(2,1,1)]] }
> king2         = king { initSoln = [[(2,1,2)]] }

Reversing a puzzle:

> reversePuzzle  :: Puzzle -> Puzzle
> reversePuzzle p = p { sections = reverse (sections p) }

Flattening a puzzle:

> flat  :: Puzzle -> Puzzle
> flat p = p { valid = (\ (x,y,z) -> z==1) }

-------------------------------------------------------------------
Solving Puzzles:

> extend :: Puzzle -> Solution -> Direction -> Int -> [Solution]
> extend p soln dir len = [ sect:soln | let start = head (head soln)
>                                           sect  = section start dir len,
>                                       all (valid p) sect,
>                                       all (`notElem` concat soln) sect ]

> solutions :: Puzzle -> [Solution]
> solutions p = solve (initSoln p) (initDir p) (sections p)
>  where solve :: Solution -> Direction -> [Int] -> [Solution]
>        solve soln dir [] = [soln]
>        solve soln dir (len:lens)
>          = concat [ solve soln' newdir lens
>                   | newdir <- newDirs dir,
>                     soln'  <- extend p soln newdir len ]

> {-
> -- Readers familiar with Haskell's do notation may prefer to write
> -- the second clause in the definition for solve as follows, instead
> -- of using a list comprehension:
>        solve soln dir (len:lens)
>               = do newdir <- newDirs dir
>                    soln'  <- extend p soln newdir len
>                    soln'' <- solve soln' newdir lens
>                    return soln''
> -}

-------------------------------------------------------------------
Leveraging Symmetry to Eliminate Solutions:

> advance  :: Puzzle -> Puzzle
> advance p = head [ p { initDir  = newdir,
>                        initSoln = soln',
>                        sections = tail (sections p) }
>                  | newdir <- newDirs (initDir p),
>                    soln'  <- extend p (initSoln p) newdir (head (sections p)) ]

-------------------------------------------------------------------
