-------------------------------------------------------------------
This file contains Haskell code for drawing configurations of the
Snake Cube Puzzle.  It is an accompaniment to following paper,
which is to be published in the Journal of Functional Programming.

             Solving the Snake Cube Puzzle in Haskell

                          MARK P. JONES
                  Department of Computer Science
          Portland State University, Portland, Oregon, USA
                         mpj@cs.pdx.edu

Please refer to the paper for further explanation of the code that
is presented here.
-------------------------------------------------------------------

> module SnakeDraw where

> import System(system)
> import Snake

Rather than generating the graphics directly, our code for drawing
configurations of the Snake Cube puzzle leverages Gene Ressler's
sketch tool, which is available online

    from http://www.frontiernet.net/~eugene.ressler/

This means that we can leave all the interesting details of
drawing a representation of a 3D object---including perspective
transforms and hidden surface removal---to sketch.  All that
remains on the Haskell side is to convert the data structures that
describe a Snake Cube configuration into a text string that uses
the syntax of sketch to provide a corresponding 3D scene
description.

Each individual position within a Snake Cube puzzle corresponds to
a cube with six faces, all of which are drawn with the same color:

> cube2sk      :: String -> Position -> [String]
> cube2sk col a = [ prefix ++ concat (map show f) | f <- faces ]
>  where prefix = "polygon[fill=" ++ col ++ "]"
>        faces  = [ [ a, d, g, c ], [ b, e, h, f ], [ a, b, f, d ],
>                   [ c, g, h, e ], [ a, c, e, b ], [ d, f, h, g ] ]
>        (x,y,z) = a
>        b = (x,y,z-1)
>        c = (x,y-1,z)
>        d = (x-1,y,z)
>        e = (x,y-1,z-1)
>        f = (x-1,y,z-1)
>        g = (x-1,y-1,z)
>        h = (x-1,y-1,z-1)

We include all six faces in our 3D scene description; sketch will
determine which of those faces can be seen, and which are
obscured, depending on the coordinates and direction from which
the scene is viewed.

Using cube2sk, it is easy to generate a sketch description for a
sequence of cubes, alternating between brown and white cube colors
to match the pattern seen in many wooden versions of the Snake
Cube puzzle.  The showCubes function defined below takes a list of
(small cube) positions as its input and produces a list of strings
as its result, each of which contains a single sketch statement.

> showCubes  :: [Position] -> [String]
> showCubes   = concat . zipWith cube2sk (cycle ["brown","white"])

This leaves us with two remaining tasks, the first of which is to
convert a given solution to a snake cube puzzle (represented by a
list of puzzle sections) into a list of puzzle configurations,
with one configuration for each step in the solution.  For this,
we can use a scan operation and some simple list processing
functions:

> steps :: Solution -> [[Position]]
> steps  = tail . reverse . map reverse . scanr1 (++)

The second task is to package up and use the output from showCubes
to generate the appropriate graphic.  The following function
accomplishes this, taking a prefix to specify the name of the
generated output file and associated temporaries as well as the
list of small cube positions.  Note that this function uses the
system function from the standard Haskell library to invoke the
sketch and pdflatex tools as well as the rm command to delete some
of the generated temporary files.  As a result, this function will
not work correctly on machines that do not have these tools
installed.

> sketch :: String -> [Position] -> IO ()
> sketch prefix config
>  = do writeFile (prefix ++ ".sk")
>                 (unlines (prologue ++ showCubes config ++ epilogue))
>       system ("sketch -T -o " ++ prefix ++ ".tex " ++ prefix ++ ".sk")
>       system ("pdflatex " ++ prefix)
>       system ("rm " ++ prefix ++ ".tex")
>       system ("rm " ++ prefix ++ ".aux")
>       system ("rm " ++ prefix ++ ".log")
>       system ("rm " ++ prefix ++ ".sk")
>       return ()

The above definition uses a prologue, which is a fixed block of
text that appears at the start of each generated sketch file.
Specifically, the prologue provides code to set an appropriate
camera/view position, and draws a grid of dotted lines to show the
positions for the small cubes in the completed puzzle.

> prologue
>  = [ "put { view((6,6,12), (0,1,-10)) then perspective(20) } {",
>      "repeat {4, translate([0,0,1])}",
>      "   {repeat {4, translate([0,1,0])} {line[style=dotted](0,0,0)(3,0,0)}}",
>      "repeat {4, translate([0,0,1])}",
>      "   {repeat {4, translate([1,0,0])} {line[style=dotted](0,0,0)(0,3,0)}}",
>      "repeat {4, translate([0,1,0])}",
>      "   {repeat {4, translate([1,0,0])} {line[style=dotted](0,0,0)(0,0,3)}}" ]

There is a corresponding epilogue, which finishes off the sketch
output description by specifying some display and output options.

> epilogue
>  = [ "}",
>      "global { set[scale=2,style=very thin] language tikz }" ]

Finally, we put everything together with a function that generates
a sequence of diagrams, using the given filename prefix, for the
first generated solution of a given puzzle.

> solutionFor :: String -> Puzzle -> IO ()
> solutionFor prefix puzzle
>  = sequence_
>  $ zipWith sketch
>            [ prefix ++ "-" ++ show n | n <- [1..] ]
>            (steps $ head $ solutions $ puzzle)

For example, if the necessary tools are installed, then

   solutionFor "standard" standard

will produce a sequence of 17 pdf files called standard-1.pdf,
standard-2.pdf, etc... that illustrate a solution to the standard
snake cube puzzle that was defined in Snake.lhs.

-------------------------------------------------------------------
